"use strict";

Object.defineProperty(exports, "__esModule", {
    value: true
});
exports.Uploader = exports.S3Client = undefined;

var _bluebirdLst;

function _load_bluebirdLst() {
    return _bluebirdLst = require("bluebird-lst");
}

var _bluebirdLst2;

function _load_bluebirdLst2() {
    return _bluebirdLst2 = _interopRequireDefault(require("bluebird-lst"));
}

var _awsSdk;

function _load_awsSdk() {
    return _awsSdk = require("aws-sdk");
}

var _events;

function _load_events() {
    return _events = require("events");
}

var _fsExtraP;

function _load_fsExtraP() {
    return _fsExtraP = require("fs-extra-p");
}

var _mime;

function _load_mime() {
    return _mime = _interopRequireDefault(require("mime"));
}

var _fdSlicer;

function _load_fdSlicer() {
    return _fdSlicer = require("./fdSlicer");
}

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const MultipartETag = require("../s3-client/multipart_etag");
const MAX_PUT_OBJECT_SIZE = 5 * 1024 * 1024 * 1024;
const MAX_MULTIPART_COUNT = 10000;
const MIN_MULTIPART_SIZE = 5 * 1024 * 1024;
(_awsSdk || _load_awsSdk()).config.setPromisesDependency(require("bluebird-lst"));
class S3Client {
    constructor() {
        let options = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

        this.s3 = options.s3Client || new (_awsSdk || _load_awsSdk()).S3(options.s3Options);
        this.s3RetryCount = options.s3RetryCount || 3;
        this.s3RetryDelay = options.s3RetryDelay || 1000;
        this.multipartUploadThreshold = options.multipartUploadThreshold || 20 * 1024 * 1024;
        this.multipartUploadSize = options.multipartUploadSize || 15 * 1024 * 1024;
        this.multipartDownloadThreshold = options.multipartDownloadThreshold || 20 * 1024 * 1024;
        this.multipartDownloadSize = options.multipartDownloadSize || 15 * 1024 * 1024;
        if (this.multipartUploadThreshold < MIN_MULTIPART_SIZE) {
            throw new Error("Minimum multipartUploadThreshold is 5MB.");
        }
        if (this.multipartUploadThreshold > MAX_PUT_OBJECT_SIZE) {
            throw new Error("Maximum multipartUploadThreshold is 5GB.");
        }
        if (this.multipartUploadSize < MIN_MULTIPART_SIZE) {
            throw new Error("Minimum multipartUploadSize is 5MB.");
        }
        if (this.multipartUploadSize > MAX_PUT_OBJECT_SIZE) {
            throw new Error("Maximum multipartUploadSize is 5GB.");
        }
    }
    createFileUploader(localFile, target, s3Options) {
        return new Uploader(this, Object.assign({ Key: target }, s3Options), localFile);
    }
}
exports.S3Client = S3Client;
class Uploader extends (_events || _load_events()).EventEmitter {
    constructor(client, s3Options, localFile) {
        super();
        this.client = client;
        this.s3Options = s3Options;
        this.localFile = localFile;
        /** @readonly */
        this.progressAmount = 0;
        /** @readonly */
        this.progressTotal = 0;
        this.cancelled = false;
        this.parts = [];
    }
    openFile() {
        var _this = this;

        return (0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* () {
            const fd = yield (0, (_fsExtraP || _load_fsExtraP()).open)(_this.localFile, "r");
            _this.fileStat = yield (0, (_fsExtraP || _load_fsExtraP()).fstat)(fd);
            _this.progressTotal = _this.fileStat.size;
            _this.fileSlicer = new (_fdSlicer || _load_fdSlicer()).FdSlicer(fd);
            const localFileSlicer = _this.fileSlicer;
            localFileSlicer.on("error", function (error) {
                _this.cancelled = true;
                _this.slicerError = error;
            });
            localFileSlicer.on("close", function () {
                _this.emit("fileClosed");
            });
            _this.emit("fileOpened", localFileSlicer);
        })();
    }
    closeFile() {
        var _this2 = this;

        return (0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* () {
            _this2.cancelled = true;
            if (_this2.fileSlicer != null) {
                _this2.fileSlicer.close();
                _this2.fileSlicer = null;
            }
        })();
    }
    upload() {
        var _this3 = this;

        return this._upload().then((() => {
            var _ref = (0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* (it) {
                if (_this3.slicerError != null) {
                    throw _this3.slicerError;
                }
                yield _this3.closeFile();
                return it;
            });

            return function (_x2) {
                return _ref.apply(this, arguments);
            };
        })()).catch((() => {
            var _ref2 = (0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* (error) {
                try {
                    yield _this3.closeFile();
                } catch (ignored) {}
                throw error;
            });

            return function (_x3) {
                return _ref2.apply(this, arguments);
            };
        })());
    }
    _upload() {
        var _this4 = this;

        return (0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* () {
            yield _this4.openFile();
            const client = _this4.client;
            if (_this4.fileStat.size >= client.multipartUploadThreshold) {
                let multipartUploadSize = client.multipartUploadSize;
                const partsRequiredCount = Math.ceil(_this4.fileStat.size / multipartUploadSize);
                if (partsRequiredCount > MAX_MULTIPART_COUNT) {
                    multipartUploadSize = smallestPartSizeFromFileSize(_this4.fileStat.size);
                }
                if (multipartUploadSize > MAX_PUT_OBJECT_SIZE) {
                    throw new Error(`File size exceeds maximum object size: ${_this4.localFile}`);
                }
                const data = yield _this4.runOrRetry(function () {
                    return client.s3.createMultipartUpload(Object.assign({ ContentType: (_mime || _load_mime()).default.lookup(_this4.localFile) }, _this4.s3Options)).promise();
                });
                yield _this4.multipartUpload(data.UploadId, multipartUploadSize);
            } else {
                yield _this4.runOrRetry(_this4.putObject.bind(_this4));
            }
        })();
    }
    abort() {
        this.cancelled = true;
    }
    putObject() {
        var _this5 = this;

        return (0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* () {
            const multipartETag = new MultipartETag({ size: _this5.fileStat.size, count: 1 });
            const multipartPromise = new (_bluebirdLst2 || _load_bluebirdLst2()).default(function (resolve, reject) {
                multipartETag.on("end", resolve);
                const inStream = (0, (_fsExtraP || _load_fsExtraP()).createReadStream)(_this5.localFile, { fd: _this5.fileSlicer.fd, autoClose: false });
                inStream.on("error", reject);
                inStream.pipe(multipartETag);
            }).then(function () {
                _this5.progressAmount = multipartETag.bytes;
                _this5.progressTotal = multipartETag.bytes;
                _this5.fileStat.size = multipartETag.bytes;
                _this5.fileStat.multipartETag = multipartETag;
                _this5.emit("progress");
            });
            multipartETag.on("progress", function () {
                if (!_this5.cancelled) {
                    _this5.progressAmount = multipartETag.bytes;
                    _this5.emit("progress");
                }
            });
            _this5.progressAmount = 0;
            const data = (yield (_bluebirdLst2 || _load_bluebirdLst2()).default.all([multipartPromise, _this5.client.s3.putObject(Object.assign({
                ContentType: (_mime || _load_mime()).default.lookup(_this5.localFile),
                ContentLength: _this5.fileStat.size,
                Body: multipartETag
            }, _this5.s3Options)).promise()]))[1];
            if (!compareMultipartETag(data.ETag, _this5.fileStat.multipartETag)) {
                throw new Error("ETag does not match MD5 checksum");
            }
        })();
    }
    multipartUpload(uploadId, multipartUploadSize) {
        var _this6 = this;

        return (0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* () {
            let cursor = 0;
            let nextPartNumber = 1;
            const parts = [];
            while (cursor < _this6.fileStat.size) {
                const start = cursor;
                let end = cursor + multipartUploadSize;
                if (end > _this6.fileStat.size) {
                    end = _this6.fileStat.size;
                }
                cursor = end;
                const part = {
                    ETag: null,
                    PartNumber: nextPartNumber++
                };
                _this6.parts.push(part);
                parts.push({ start, end, part });
            }
            yield (_bluebirdLst2 || _load_bluebirdLst2()).default.map(parts, function (it) {
                return _this6.makeUploadPart(it, uploadId);
            }, { concurrency: 4 });
            return yield _this6.runOrRetry(function () {
                return _this6.client.s3.completeMultipartUpload({
                    Bucket: _this6.s3Options.Bucket,
                    Key: _this6.s3Options.Key,
                    UploadId: uploadId,
                    MultipartUpload: {
                        Parts: _this6.parts
                    }
                }).promise();
            });
        })();
    }
    makeUploadPart(p, uploadId) {
        var _this7 = this;

        return this.runOrRetry((0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* () {
            const client = _this7.client;
            let errorOccurred = false;
            const contentLength = p.end - p.start;
            const multipartETag = new MultipartETag({ size: contentLength, count: 1 });
            let prevBytes = 0;
            let overallDelta = 0;
            multipartETag.on("progress", function () {
                if (_this7.cancelled || errorOccurred) {
                    return;
                }
                const delta = multipartETag.bytes - prevBytes;
                prevBytes = multipartETag.bytes;
                _this7.progressAmount += delta;
                overallDelta += delta;
                _this7.emit("progress");
            });
            const inStream = new (_fdSlicer || _load_fdSlicer()).SlicedReadStream(_this7.fileSlicer, p.start, p.end);
            const multipartPromise = new (_bluebirdLst2 || _load_bluebirdLst2()).default(function (resolve, reject) {
                inStream.on("error", reject);
                multipartETag.on("end", resolve);
            }).then(function () {
                if (_this7.cancelled || errorOccurred) {
                    return;
                }
                _this7.progressAmount += multipartETag.bytes - prevBytes;
                _this7.progressTotal += p.end - p.start - multipartETag.bytes;
                _this7.emit("progress");
            });
            inStream.pipe(multipartETag);
            const data = (yield (_bluebirdLst2 || _load_bluebirdLst2()).default.all([multipartPromise, client.s3.uploadPart({
                ContentLength: contentLength,
                PartNumber: p.part.PartNumber,
                UploadId: uploadId,
                Body: multipartETag,
                Bucket: _this7.s3Options.Bucket,
                Key: _this7.s3Options.Key
            }).promise().catch(function (error) {
                errorOccurred = true;
                _this7.progressAmount -= overallDelta;
                throw error;
            })]))[1];
            if (!compareMultipartETag(data.ETag, multipartETag)) {
                _this7.progressAmount -= overallDelta;
                throw new Error("ETag does not match MD5 checksum");
            }
            p.part.ETag = data.ETag;
            return data;
        }));
    }
    runOrRetry(task) {
        var _this8 = this;

        return (0, (_bluebirdLst || _load_bluebirdLst()).coroutine)(function* () {
            return new (_bluebirdLst2 || _load_bluebirdLst2()).default(function (resolve, reject) {
                let attemptNumber = 0;
                const tryRun = function () {
                    if (_this8.cancelled) {
                        return;
                    }
                    task().then(resolve).catch(function (error) {
                        if (++attemptNumber >= _this8.client.s3RetryCount) {
                            reject(error);
                        } else if (_this8.cancelled) {
                            reject(new Error("cancelled"));
                        } else {
                            setTimeout(tryRun, _this8.client.s3RetryDelay);
                        }
                    });
                };
                tryRun();
            });
        })();
    }
}
exports.Uploader = Uploader;
function cleanETag(eTag) {
    return eTag == null ? "" : eTag.replace(/^\s*'?\s*"?\s*(.*?)\s*"?\s*'?\s*$/, "$1");
}
function compareMultipartETag(eTag, multipartETag) {
    return multipartETag.anyMatch(cleanETag(eTag));
}
function smallestPartSizeFromFileSize(fileSize) {
    const partSize = Math.ceil(fileSize / MAX_MULTIPART_COUNT);
    return partSize < MIN_MULTIPART_SIZE ? MIN_MULTIPART_SIZE : partSize;
}
//# sourceMappingURL=uploader.js.map